from datetime import datetime, date
from typing import Any

from sanic.request import File

from insanic.utils import Container, ContainerField


class FieldValidationError(Exception):
    def __init__(self, error_code: str):
        super().__init__()
        self.error_code = error_code

class Field(ContainerField):
    ERROR_REQUIRED = 'required'

    def __init__(self, *args: Any, required: bool = True, default: Any | None = None, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.required = required
        if self.required and default:
            raise UserWarning('Cannot set default for required field')

        self.default = default

    def validate_value(self, value: Any) -> Any | None: # pylint: disable=no-self-use
        return value # pragma: no cover

    def validate(self, value: Any) -> Any | None:
        if value:
            return self.validate_value(value)

        if self.required:
            raise FieldValidationError(self.ERROR_REQUIRED)

        return self.default

class FileField(Field):
    ERROR_INVALID_FILE_TYPE = 'invalid_file_type'

    allowed_file_types: list[str]

    def __init__(self, *args: Any, file_type: str | list[str] | None = None, **kwargs: Any) -> None:
        self.allowed_file_types = [file_type] if isinstance(file_type, str) else file_type
        super().__init__(*args, **kwargs)

    def validate_value(self, value: File) -> File:
        if value.type not in self.allowed_file_types:
            raise FieldValidationError(self.ERROR_INVALID_FILE_TYPE)

        return value

class IntegerField(Field):
    ERROR_NOT_INTEGER = 'invalid_integer'

    def validate_value(self, value) -> None:
        try:
            return int(value)
        except ValueError as error:
            raise FieldValidationError(self.ERROR_NOT_INTEGER) from error

class DateTimeField(Field):
    ERROR_INVALID_DATE = 'invalid_date'

    def __init__(self, *args, parse_format='%Y-%m-%d %H:%M:%S', **kwargs) -> None:
        self.parse_format = parse_format
        super().__init__(*args, **kwargs)

    def validate_value(self, value) -> datetime:
        try:
            return datetime.strptime(value, self.parse_format)
        except ValueError as error:
            raise FieldValidationError(self.ERROR_INVALID_DATE) from error

class DateField(DateTimeField):
    def __init__(self, *args, **kwargs) -> None:
        kwargs['parse_format'] = kwargs.pop('parse_format', '%Y-%m-%d')
        super().__init__(*args, **kwargs)

    def validate_value(self, value) -> date:
        value = super().validate_value(value)
        return value.date()



from typing import Any

from insanic.commands import SyncCommand, CommandParser
from insanic.utils import load_application

class ServerCommand(SyncCommand):
    help = 'Runs application server'

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('-p', help='Server port')

    def execute(self, **kwargs: Any) -> None:
        application = load_application()

        server_config = {
            'host': application.config.APPLICATION_HOST,
            'port': application.config.APPLICATION_PORT,

            'debug': application.config.APPLICATION_DEBUG,
            'auto_reload': application.config.APPLICATION_DEBUG,
            'access_log': application.config.APPLICATION_DEBUG,

            'workers': application.config.APPLICATION_WORKERS,
        }
        application.run(**server_config)



import os
import warnings
from pathlib import Path
import sys
from typing import Any

from insanic.db import get_db_connection, in_transaction
from insanic.commands import AsyncCommand, SyncCommand, CommandParser
from insanic.utils import read_file_content

MIGRATIONS_TABLE = 'db_migrations'
MIGRATIONS_LOCATION = Path('migrations')

MIGRATION_UP_MARKER = '-- upgrade --'
MIGRATION_DOWN_MARKER = '-- downgrade --'
MIGRATIONS_TEMPLATE = f"""
{MIGRATION_UP_MARKER}

{MIGRATION_DOWN_MARKER}
"""

class Migration:
    STATE_APPLIED: str = '*'
    STATE_PLANNED: str = ' '
    STATE_MISSING: str = '?'

    name: str
    state: str | None

    def __init__(self, name: str, state: str | None = None) -> None:
        self.name = name
        self.state = state

async def create_migrations_table() -> None:
    connection = get_db_connection()

    # @FIXME: Ugly hack to prevent "Warning: Table 'db_migrations' already exists"
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        await connection.execute_query(f'''
            CREATE TABLE IF NOT EXISTS `{MIGRATIONS_TABLE}` (
                `name` varchar(255) NOT NULL,
                `upgrade_sql` text NOT NULL,
                `downgrade_sql` text NULL DEFAULT NULL,

                PRIMARY KEY (`name`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_0900_ai_ci
        ''')

async def get_applied_migrations() -> list[str]:
    connection = get_db_connection()
    await create_migrations_table()

    _, query_result = await connection.execute_query(f'SELECT `name` FROM `{MIGRATIONS_TABLE}`')
    return [ row['name'] for row in query_result ]

def get_migrations_files() -> list[str]:
    return list(sorted(filter(lambda x: x.endswith('sql'), os.listdir(MIGRATIONS_LOCATION))))

async def get_migrations_list() -> list[Migration]:
    applied_migrations = await get_applied_migrations()
    migrations_files = get_migrations_files()

    migrations = []
    for applied_migration in applied_migrations:
        migration = Migration(applied_migration, state=Migration.STATE_MISSING)

        if applied_migration in migrations_files:
            migration.state = Migration.STATE_APPLIED
            migrations_files.remove(applied_migration)

        migrations.append(migration)

    for migration_file in migrations_files:
        migrations.append(Migration(migration_file, state=Migration.STATE_PLANNED))

    return sorted(migrations, key=lambda migration: migration.name)

async def apply_migration_file(file_name: str) -> None:
    migration_instructions = read_file_content(Path(MIGRATIONS_LOCATION, file_name))

    upgrade_queries_index = migration_instructions.index(MIGRATION_UP_MARKER) + len(MIGRATION_UP_MARKER)
    try:
        downgrade_queries_index = migration_instructions.index(MIGRATION_DOWN_MARKER) + len(MIGRATION_DOWN_MARKER)
    except ValueError:
        downgrade_queries_index = len(migration_instructions) - 1

    upgrade_queries = migration_instructions[upgrade_queries_index:downgrade_queries_index].strip()
    downgrade_queries = migration_instructions[downgrade_queries_index:].strip()

    async with in_transaction() as connection:
        await connection.execute_script(upgrade_queries)
        await connection.execute_insert(f'INSERT INTO `{MIGRATIONS_TABLE}` VALUES (%s, %s, %s)', [
            file_name, upgrade_queries, downgrade_queries,
        ])

async def rollback_migration(migration_name: str) -> None:
    connection = get_db_connection()

    migrations_list = await connection.execute_query_dict(
        f'SELECT * FROM `{MIGRATIONS_TABLE}` WHERE `name` = %s LIMIT 1',
        [migration_name]
    )
    migration = migrations_list[0]

    async with in_transaction() as connection:
        await connection.execute_script(migration['downgrade_sql'])
        await connection.execute_insert(f'DELETE FROM `{MIGRATIONS_TABLE}` WHERE `name` = %s', [migration_name])


class MigrationsListCommand(AsyncCommand):
    help = 'Lists all available migrations'

    async def execute(self, **kwargs: Any) -> None:
        migrations = await get_migrations_list()

        print('List of migrations:')
        print('Status | Migration')

        for migration in migrations:
            print(f'     {migration.state} | {migration.name}')

class MigrationsApplyCommand(AsyncCommand):
    help = 'Applies all planned migrations'

    async def execute(self, **kwargs: Any) -> None:
        migrations = await get_migrations_list()
        planned_migrations_count = len(list(filter(lambda x: x.state == Migration.STATE_PLANNED, migrations)))

        print(f'Planned migrations count: {planned_migrations_count}')

        for migration in migrations:
            if migration.state == Migration.STATE_APPLIED:
                continue

            print(f'Migration "{migration.name}": ', end='')

            if migration.state == Migration.STATE_PLANNED:
                print('Applying... ', end='')
                sys.stdout.flush()

                await apply_migration_file(migration.name)
                print('OK')
                continue

            if migration.state == Migration.STATE_MISSING:
                print('Rolling back... ', end='')
                sys.stdout.flush()

                await rollback_migration(migration.name)
                print('OK')
                continue

class MigrationsGenerateCommand(SyncCommand):
    help = 'Generates a new migration file'

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('filename', nargs='?', default='migration')

    def generate_migration_number(self) -> int:  # pylint: disable=no-self-use
        migration_files = get_migrations_files()

        if len(migration_files) == 0:
            return 0

        return int(migration_files[-1][0:4])

    def execute(self, filename: str, **kwargs: Any) -> None:  # type: ignore[override] # pylint: disable=arguments-differ
        migration_number = self.generate_migration_number() + 1
        migration_filename = Path(MIGRATIONS_LOCATION, f'{migration_number:04d}_{filename}.sql')

        with open(migration_filename, 'wt', encoding='utf-8') as file:
            file.write(MIGRATIONS_TEMPLATE)

        print(f'Migration file "{migration_filename} is generated...')



from typing import Any

from insanic.utils import Container

from .fields import Field, FieldValidationError


class FormValidationError(Exception):
    def __init__(self, field_errors: dict[str, FieldValidationError]):
        super().__init__()
        self.errors = field_errors


FormData = dict[str, Any]
FormErrors = dict[str, FieldValidationError]
class Form(Container):
    FIELD_BASE_CLASS = Field
    RESERVED_ATTRS = ['is_validated', 'errors', 'data']

    def __init__(self, data: FormData | object, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._raw_data: FormData | object = data
        self._data: FormData = {}
        self._errors: FormErrors = {}

    @property
    def is_validated(self) -> bool:
        return len(self._data.keys()) > 0 or len(self._errors.keys()) > 0

    @property
    def errors(self) -> FormErrors:
        assert self.is_validated, 'Form hasn\'t been validated'
        return self._errors

    @property
    def data(self) -> FormData:
        assert self.is_validated, 'Form hasn\'t been validated'
        return self._data

    def add_error(self, field_name: str, error: FieldValidationError) -> None:
        self.errors[field_name] = error

    async def run_validation(self) -> None:
        pass

    async def validate(self) -> FormData:
        self._data = {}
        self._errors = {}

        for field_name, field_value, field in self._field_values(self._raw_data):
            try:
                self._data[field_name] = field.validate(field_value)
            except FieldValidationError as validation_error:
                self._errors[field_name] = validation_error

        await self.run_validation()

        if len(self._errors) > 0:
            self._data = {}
            raise FormValidationError(self._errors)

        return self.data

    async def is_valid(self) -> bool:
        if not self.is_validated:
            try:
                await self.validate()
            except FormValidationError as _exc:
                pass

        return len(self._errors) == 0



import logging
import sys
from argparse import ArgumentParser
from typing import Any

from insanic.utils import application_context

class CommandParser(ArgumentParser):
    pass

class Command:
    help: str | None = None
    parser: CommandParser = CommandParser()

    def exit(self, exit_code: int = 0) -> None:  # pylint: disable=no-self-use
        sys.exit(exit_code)

    def add_arguments(self, parser: CommandParser) -> None:
        pass

class SyncCommand(Command):
    def execute(self, **kwargs: Any) -> None:
        raise NotImplementedError()

class AsyncCommand(Command):
    async def execute_with_context(self, **kwargs: Any) -> None:
        async with application_context():
            await self.execute(**kwargs)

    async def execute(self, **kwargs: Any) -> None:
        raise NotImplementedError()



import os
import sys
from typing import cast

from insanic.commands import (
    AsyncCommand,
    CommandParser,
    SyncCommand,
)
from insanic.commands.migrations import (
    MigrationsApplyCommand,
    MigrationsGenerateCommand,
    MigrationsListCommand,
)
from insanic.commands.routes import RoutesCommand
from insanic.commands.server import ServerCommand
from insanic.commands.tasks import TasksQueueCommand
from insanic.utils import run_async

# @TODO: Implement dynamic commands loading
def load_commands() -> dict[str, AsyncCommand | SyncCommand]:
    return {
        'routes': RoutesCommand(),
        'server': ServerCommand(),

        'tasks_queue': TasksQueueCommand(),

        'migrations:apply': MigrationsApplyCommand(),
        'migrations:generate': MigrationsGenerateCommand(),
        'migrations:list': MigrationsListCommand(),
    }

def execute() -> None:
    # Adding `./src` into sys path to be able to import relative modules
    app_path = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), 'src')
    sys.path = [app_path] + sys.path

    parser = CommandParser(prog='Insanic CLI')
    subparsers = parser.add_subparsers(dest='command_name')

    commands = load_commands()

    # @TODO: Sort commands
    for name, command in commands.items():
        subparser = subparsers.add_parser(name, help=command.help)
        command.add_arguments(cast(CommandParser, subparser))

    cli_args = parser.parse_args().__dict__

    command_name = cli_args.pop('command_name', None)
    if not command_name:
        parser.print_help()
        sys.exit()

    # @TODO: Handle error
    command = commands[command_name]
    if isinstance(command, AsyncCommand):
        run_async(command.execute_with_context(**cli_args))
    else:
        command.execute(**cli_args)



from __future__ import annotations

import importlib
from asyncio import AbstractEventLoop
from typing import Any

from sanic import Sanic

from insanic.config import DEFAULT_CONFIG
from insanic.db import init_db_connection, close_db_connection
from insanic.http import Request
from insanic.tasks import init_redis_pool, close_redis_pool

class Application(Sanic):
    def __init__(self, *args: Any, **kwargs: Any):
        # @TODO: Implement custom error handler
        # @TODO: Implement proper logging

        # Overriding Sanic default implementations
        kwargs['strict_slashes'] = True
        kwargs['request_class'] = Request  # Setting up custom Request class
        super().__init__(*args, **kwargs)

        self.load_configuration()

        self.register_listener(self.boot, 'before_server_start')
        self.register_listener(self.shutdown, 'after_server_stop')

    def load_configuration(self) -> None:
        # Default configuration
        self.config.update_config(DEFAULT_CONFIG)

        # Application configuration
        try:
            application_config = importlib.import_module('config', package='src') # importing src/config.py
            self.config.update_config(application_config)
        except ModuleNotFoundError as _error:
            pass  # @TODO: Log that application is using default config

        try:
            local_config = importlib.import_module('config_local', package='src')  # importing src/config_local.py
            self.config.update_config(local_config)
            # @TODO: Log that application is using local config
        except ModuleNotFoundError as _error:
            pass

        # Applying ENV config
        self.config.load_environment_vars(prefix='INSANIC_')

    async def boot(self, _application: Application, _loop: AbstractEventLoop) -> None: # pylint: disable=no-self-use
        await init_db_connection()
        await init_redis_pool()

    async def shutdown(self, _application: Application, _loop: AbstractEventLoop) -> None: # pylint: disable=no-self-use
        await close_db_connection()
        await close_redis_pool()



from typing import Any

from sanic.router import Route

from insanic.app import Application
from insanic.commands import SyncCommand, CommandParser
from insanic.tasks import TaskDefinition, run_workers
from insanic.utils import load_application


class TasksQueueCommand(SyncCommand):
    help = 'Runs workers for specified task'

    def print_tasks_list(self, available_tasks) -> None:
        for task_name, task_defintion in available_tasks.items():
            print(f'{task_name}\tworkers: {task_defintion.workers}, poll_delay: {task_defintion.poll_delay}')

    def get_available_tasks(self, application: Application) -> dict[str, TaskDefinition]:
        available_tasks = {}

        for blueprint_name, blueprint in application.blueprints.items():
            for task_name, task_definition in blueprint.tasks.items():
                available_tasks[f'{blueprint_name}:{task_name}'] = task_definition

        return available_tasks

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('task_name', nargs='?')

    def execute(self, task_name: str | None, **kwargs: Any) -> None:
        application = load_application()

        available_tasks = self.get_available_tasks(application)

        if not task_name:
            return self.print_tasks_list(available_tasks)

        task_definition = available_tasks.get(task_name, None)
        if not task_definition:
            raise Exception(f'Task "{task_name}" is not registered')

        run_workers(task_name, task_definition)



from typing import Any

from sanic.views import HTTPMethodView

from insanic.http import HTTP_STATUS, Response, response_json
from insanic.serializers import Serializer

__all__ = (
    'View',
)


class View(HTTPMethodView):
    serializer_class: type[Serializer] | None = None

    def ok_response(self, data: Any, http_status: int = HTTP_STATUS.OK) -> Response:  # pylint: disable=no-self-use
        response_data = { 'status': 'ok', 'data': data }
        return response_json(response_data, status=http_status)

    def error_response(self, errors: dict, http_status : int = HTTP_STATUS.BAD_REQUEST) -> Response:  # pylint: disable=no-self-use
        response_data = { 'status': 'error', 'errors': errors }
        return response_json(response_data, status=http_status)

    async def serialized_response(self, data: Any, serializer_class: type[Serializer] | None = None, http_status: int = HTTP_STATUS.OK) -> Response :  # pylint: disable=line-too-long
        serializer_class = serializer_class or self.serializer_class
        if not serializer_class:
            raise NotImplementedError('Serializer class must be specified')

        serializer_kwargs = { 'many': isinstance(data, list) }
        serializer = serializer_class(**serializer_kwargs)
        serialized_data = await serializer.serialize(data)

        return self.ok_response(serialized_data, http_status=http_status)



from typing import Any

from sanic.router import Route

from insanic.commands import SyncCommand
from insanic.utils import load_application

def sort_routes(route: Route) -> str:
    is_app_route = len(route.segments) != 1
    route_first_letter = route.segments[0][0]  # First letter of first segments

    return (is_app_route, route_first_letter,)

class RoutesCommand(SyncCommand):
    help = 'Prints all registered routes'

    def execute(self, **kwargs: Any) -> None:
        application = load_application()

        sorted_routes = sorted(application.router.routes_all.values(), key=sort_routes)
        for route in sorted_routes:
            print(route.uri)



import os

from sanic import Sanic
from tortoise import Tortoise, Model, fields, functions, queryset
from tortoise.backends.base.client import BaseDBAsyncClient
from tortoise.manager import Manager as ModelManager
from tortoise.transactions import atomic as with_transaction, in_transaction


__all__ = (
    'Model',
    'ModelManager',
    'fields',
    'functions',
    'queryset',
    'in_transaction',
    'with_transaction',
    'init_db_connection',
    'close_db_connection',
)

def get_db_connection() -> BaseDBAsyncClient:
    return Tortoise.get_connection('default')

async def init_db_connection() -> BaseDBAsyncClient:
    application = Sanic.get_app()
    database_url = application.config.DATABASE_URL

    # Building list of blueprints with `models.py`
    # to add them into list of modules for TortoiseORM
    blueprint_models = {}
    for blueprint_name in application.blueprints:
        if os.path.exists(f'src/{blueprint_name}/models.py'):
            blueprint_models[blueprint_name] = [f'{blueprint_name}.models']

    # @TODO: Fix this ugly hack. Tortoise doesn't like empty modules so we trick it
    # empty_models: dict[str, dict] = {'models': {}}

    await Tortoise.init(db_url=database_url, modules=blueprint_models or None)  # type: ignore
    return get_db_connection()

async def close_db_connection() -> None:
    await Tortoise.close_connections()



"""Common classes and methods used in framework."""
import asyncio
import importlib
from collections.abc import AsyncGenerator, Coroutine, Generator
from contextlib import asynccontextmanager
from inspect import isawaitable
from pathlib import Path
from typing import Any, Awaitable, Callable, cast

from insanic.app import Application

class ContainerField:
    source: str | None

    """Common class for Form and Serializer."""
    def __init__(self, source: str | None = None):
        self.source = source

class Container:
    FIELD_BASE_CLASS = ContainerField
    RESERVED_ATTRS: list[str] = []

    _field_names: list[str] = []

    def __init__(self, *args: Any, **kwargs: Any) -> None: # pylint: disable=unused-argument
        # List of declared fields
        self._field_names = [
            attribute
            for attribute in dir(self)
            if attribute not in self.RESERVED_ATTRS and isinstance(getattr(self, attribute), self.FIELD_BASE_CLASS)
        ]

    def _field_values(self, data: Any) -> Generator:
        for field_name in self._field_names:
            field: ContainerField = getattr(self, field_name)
            field_source = getattr(field, 'source', None)  # Using field's `source` attribute as value path
            property_value_path = field_source or field_name
            field_value = get_property_value(data, property_value_path)
            yield field_name, field_value, field

def get_property_value(entity: dict | object, key: str) -> Any:
    """Get value by key from dict or any object.

    Returns None if key missing in entity
    """
    if isinstance(entity, dict):
        return entity.get(key, None)

    return getattr(entity, key, None)

def read_file_content(file_path: str | Path) -> str:
    with open(file_path, 'rt', encoding="utf-8") as file:
        return file.read()

def run_async(coroutine: Coroutine) -> Any:
    return asyncio.run(coroutine)

async def ensure_run(result: Awaitable | Callable) -> Any:
    if isawaitable(result):
        return await cast(Awaitable, result)

    return result

def load_application() -> Application:
    app_module = importlib.import_module('app', package='src')  # importing src/app.py
    application: Application = getattr(app_module, 'application')

    return application

@asynccontextmanager
async def application_context() -> AsyncGenerator:
    application = load_application()
    loop = asyncio.get_running_loop()

    await application.boot(application, loop)
    try:
        yield application
    except Exception as error:
        raise error
    finally:
        await application.shutdown(application, loop)



from http import HTTPStatus as HTTP_STATUS
from typing import Any

from sanic.request import Request as SanicRequest, File
from sanic.response import BaseHTTPResponse, HTTPResponse

__all__ = (
    'HTTP_STATUS',
    'File',
    'Request',
    'Response',
)

class Request(SanicRequest):
    pass

class Response(HTTPResponse):
    pass

def response_json(body: Any, status: int = 200, headers: dict[str, str] | None = None, **kwargs: Any) -> Response:
    json_data = BaseHTTPResponse._dumps(body, **kwargs)  # pylint: disable=protected-access
    return Response(json_data, headers=headers, status=status, content_type='application/json')



import inspect
from typing import Any, Callable

from sanic.blueprints import Blueprint as SanicBlueprint
from sanic.constants import HTTP_METHODS

from insanic.tasks import Task, TaskDefinition
from insanic.views import View

__all__ = (
    'Blueprint',
)

# Wrapper for class method routes.
# Inits class instance and executes method
# @TODO: Implement `before_method` for class based views
def wrap_class_handler(view_class: type, handler: Callable) -> Callable:
    def dispatcher(*args: Any, **kwargs: Any) -> Any:
        view = view_class()
        return handler(view, *args, **kwargs)

    # @TODO: Ugly hack and I don't remember for what!
    dispatcher.target = handler  # type: ignore[attr-defined]
    return dispatcher

class Blueprint(SanicBlueprint):  # pylint: disable=abstract-method
    tasks: dict[str, TaskDefinition] = {}

    def register_task(self, name: str, task_cls: type[Task], workers: int = 1, poll_delay: float = 0.5):
        self.tasks[name] = TaskDefinition(task_cls, workers=workers, poll_delay=poll_delay)

    def route_method(self, handler: Callable, *args: Any, http_method: str | None = None, **kwargs: Any) -> None:
        # Support for class methods routes like `ClassName.MethodName`
        # "staticmethod" stlye but with class initialization (see `wrap_class_handler`)
        if inspect.signature(handler).parameters.get('self'):
            handler_module = inspect.getmodule(handler)
            handler_class_name = handler.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
            handler_class = getattr(handler_module, handler_class_name)
            handler = wrap_class_handler(handler_class, handler)

        http_method = http_method or 'GET'
        kwargs['methods'] = [http_method]
        self.route(*args, **kwargs)(handler)

    # @FIXME: There is gotta be better way of routing the view
    def route_view(self, view_class: type[View], url: str, *args: Any, mapping: dict[str, Callable | tuple[Callable, str]] | None = None, **kwargs: Any) -> None:  # pylint: disable=line-too-long
        # @TODO: Validate view_class class
        mapping = mapping or {}
        kwargs['strict_slashes'] = kwargs.get('strict_slashes', self.strict_slashes)

        for method in HTTP_METHODS:
            mapped_handler = mapping.get(method.upper(), None)

            if isinstance(mapped_handler, tuple):
                method_url = f'{url.removesuffix("/")}/{mapped_handler[1].removeprefix("/")}'
                mapped_handler = mapped_handler[0]
            else:
                method_url = url

            handler_method = mapped_handler or getattr(view_class, method.lower(), None)

            if handler_method:  # @TODO: Validate handler_method is a method
                handler = wrap_class_handler(view_class, handler_method)
                self.route(method_url, *args, **kwargs, methods=(method,))(handler)

    def route_get(self, *args: Any, **kwargs: Any) -> None:
        kwargs['http_method'] = 'GET'
        self.route_method(*args, **kwargs)

    def route_post(self, *args: Any, **kwargs: Any) -> None:
        kwargs['http_method'] = 'POST'
        self.route_method(*args, **kwargs)



from functools import partial
from typing import Coroutine

from arq import create_pool
from arq.connections import RedisSettings
from arq.worker import Worker as ArqWorker
from sanic import Sanic


class Task:
    @classmethod
    async def enqueue(cls, *args, job_id: str | None = None, **kwargs):
        application = Sanic.get_app()
        task_name = cls.__qualname__

        await application.ctx.redis_pool.enqueue_job(task_name, *args, _job_id=job_id, **kwargs)

class TaskDefinition:
    workers: int
    poll_delay: float

    task_cls: Task

    def __init__(self, task_cls: Task, workers: int = 1, poll_delay: float = 0.5):
        self.task_cls = task_cls

        self.workers = workers
        self.poll_delay = poll_delay

async def init_redis_pool() -> None:
    application = Sanic.get_app()
    redis_url = application.config.REDIS_URL
    application.ctx.redis_pool = await create_pool(RedisSettings.from_dsn(redis_url))

async def close_redis_pool() -> None:
    #@TODO: Implement pool release
    pass

async def run_task(ctx, *args, task_cls: type[Task], **kwargs) -> None:
    task = task_cls()
    await task.execute(*args, **kwargs)

def run_workers(task_name, task_definition: TaskDefinition) -> None:
    worker_function = partial(run_task, task_cls=task_definition.task_cls)
    # Some hack to pass partial to ArqWorker (see arq/worker.py:81)
    worker_function.__qualname__ = task_definition.task_cls.__qualname__

    worker = ArqWorker(
        functions=[worker_function],
        max_jobs=task_definition.workers,
        poll_delay=task_definition.poll_delay,
        keep_result=0,
    )

    worker.run()



from typing import Any

from insanic.utils import Container

from .fields import Field


class SerializationError(Exception):
    pass

class Serializer(Container):
    FIELD_BASE_CLASS = Field

    def __new__(cls, *args, many=False, **kwargs): # type: ignore
        if many:
            # @TODO: Ignoring type due https://github.com/python/mypy/issues/6799
            return ListSerializer(serializer_cls=cls, *args, **kwargs)  # type: ignore[misc]
        return super().__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # Adding support for nested serializers
        for attribute in dir(self):
            if isinstance(getattr(self, attribute), self.__class__.__bases__ + (ListSerializer, )):
                self._field_names.append(attribute)

    async def serialize(self, data: dict) -> dict:
        return {
            field_name: await field.serialize(field_value)
            for field_name, field_value, field in self._field_values(data)
        }

class ListSerializer:
    serializer_class: type[Serializer]

    def __init__(self, serializer_cls: type[Serializer]) -> None:
        self.serializer_class = serializer_cls

    async def serialize(self, data: dict) -> list[dict]:
        serializer = self.serializer_class()

        return [
            await serializer.serialize(entity)
            for entity in data
        ]



from typing import Any

from datetime import date

from insanic.utils import (
    get_property_value,
    ContainerField,
)


__all__ = (
    'Field',
    'StringField',
)

class FieldSerializationError(Exception):
    pass

class Field(ContainerField):
    async def serialize_value(self, value: Any) -> Any: # pylint: disable=no-self-use
        return value # pragma: no cover

    async def serialize(self, value: Any) -> Any:
        return await self.serialize_value(value)

class BooleanField(Field):
    async def serialize_value(self, value: Any) -> bool | None:
        if value is None:
            return None

        return bool(value)

class IntegerField(Field):
    async def serialize_value(self, value: Any) -> int | None:
        if value is None:
            return None

        return int(value)

class FloatField(Field):
    async def serialize_value(self, value: Any) -> float | None:
        if value is None:
            return None

        return float(value)

class StringField(Field):
    async def serialize_value(self, value: Any) -> str | None:
        if value is None:
            return None

        return str(value)

class DateField(Field):
    def __init__(self, date_format: str = '%Y-%m-%d', **kwargs: Any) -> None:
        self.date_format = date_format
        super().__init__(**kwargs)

    # @TODO: Check if value a proper date?
    async def serialize_value(self, value: date | None) -> str | None:
        if value is None:
            return None

        return value.strftime(self.date_format)

class DictKeyField(Field):
    def __init__(self, key: str, **kwargs: Any) -> None:
        self.key = key
        super().__init__(**kwargs)

    async def serialize_value(self, value: dict) -> Any:
        return get_property_value(value, self.key)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['insanic', 'insanic.commands', 'insanic.forms', 'insanic.serializers']

package_data = \
{'': ['*']}

install_requires = \
['arq>=0.22,<0.23',
 'asyncmy>=0.2.3,<0.3.0',
 'cryptography>=36.0.1,<37.0.0',
 'sanic>=21.12.0,<22.0.0',
 'tortoise-orm>=0.18.0,<0.19.0']

setup_kwargs = {
    'name': 'insanic',
    'version': '0.1.6',
    'description': 'Django like web framework based on Sanic and TortoiseORM',
    'long_description': 'None',
    'author': 'Dmitry Ovchinnikov',
    'author_email': 'mail@dimka.online',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
